<?php namespace MTTPlugin\Lib\Abstracts;

/**
 * This class helps using WordPress database without custom post types
 *
 * @since 1.0.0
 */
abstract class Model {

	// MODEL DB TABLE
	/**
	 * The database table name
	 *
	 * @var 	string
	 *
	 * @since 	1.0.0
	 */
	protected static $table_name = '';

	/**
	 * Columns in the database table
	 *
	 * @var array
	 *
	 * @since    1.0.0
	 */
	protected static $columns = array();

	/**
	 * Columns extra
	 * @var array
	 */
	protected static $columns_extra = array(
		'created_date' => array(
			'type' => 'datetime'
		),
		'created_by' => array(
			'type' => 'int',
			'unsigned' => true
		),
		'updated_date' => array(
			'type' => 'datetime'
		),
		'updated_by' => array(
			'type' => 'int',
			'unsigned' => true
		)
	);

	/**
	 * Primary database table column
	 *
	 * @var string
	 *
	 * @since    1.0.0
	 */
	protected static $primary_key = 'id';

	// META DB TABLE
    /**
     * Meta table name
     *
     * @var string
     *
     * @since 5.0.0
     */
	protected static $meta_table_name = '';

    /**
     * Meta table columns
     *
     * @var array
     *
     * @since 5.0.0
     */
	protected static $meta_columns = array(
        'id' => array(
            'type' => 'bigint',
            'unsigned' => true,
            'not_null' => true
        ),
        'parent_id' => array(
            'type' => 'int',
            'unsigned' => true
        ),
        'meta_key' => array(
            'type' => 'varchar(256)'
        ),
        'meta_value' => array(
            'type' => 'longtext'
        )
    );

    /**
     * Meta Table Primary key
     *
     * @var string
     *
     * @since 5.0.0
     */
	protected static $meta_primary_key = 'id';

	// OTHER INFO
	/**
	 * Version of the database
	 *
	 * This is used for table update.
	 *
	 * @var string
	 *
	 * @since    1.0.0
	 */
	protected static $db_version = '1.0.0';

	// HELPERS
	/**
	 * Total count of last query
	 *
	 * @var    int
	 *
	 * @since    1.0.0
	 */
	protected static $total_num_rows = 0;

	/**
	 * Store runtime cache
	 *
	 * @var    array
	 *
	 * @since    1.0.0
	 */
	protected static $cache = array();


	// INITIALIZERS
	/**
	 * Model constructor.
	 *
	 * @param 	array 	$properties 	Array of properties
	 *
	 * @since 	1.0.0
	 */
	public function __construct($properties = array()) {
		$this->init($properties);

		return $this;
	}

	/**
	 * Initialize properties
	 *
	 * @param 	mixed	$properties		If is int, it will assume it is an ID, and get properties from the database
	 *
	 * @since   1.0.0
	 */
	protected function init($properties = array()) {
		$this->{static::$primary_key} = 0;

		if(is_int($properties)) {
			$properties = $this->getFirst(array(
				'where' => array(
					'id' => $properties,
				),
				'return' => 'ARRAY_A',
			));
		}
		if(is_array($properties)) {
			foreach($properties as $key => $property) {
				$this->{$key} = $property;
			}
		}
	}

    /**
     * Register Model
     *
     * @since   5.0.0
     */
	public static function register() {
		static::createTable();
		static::createMetaTable();
		static::registerModelActions();
		static::registerModelFilters();
	}

	/**
	 * Register Actions specified for the Model
	 *
	 * @since 5.0.0
	 */
	protected static function registerModelActions() {

	}

	/**
	 * Register Filters specified for the Model
	 *
	 * @since 5.0.0
	 */
	protected static function registerModelFilters() {

	}

	// SETTERS

	// GETTERS
	/**
	 * Generates slug from class name
	 *
	 * @return string
	 * @throws \ReflectionException
	 */
	public static function getSlug() {
		return strtolower((new \ReflectionClass(__CLASS__))->getShortName());
	}

	/**
	 * Returns table name with WordPress databses prefix
	 *
	 * @return 	string 	Table name
	 *
	 * @since	1.0.0
	 */
	public static function getTableName() {
		global $wpdb;

		return $wpdb->prefix . static::$table_name;
	}

	/**
	 * Returns the meta table name withs WordPress database prefix
	 * @return string
	 */
	public static function getMetaTableName() {
		global $wpdb;

		return $wpdb->prefix . static::$meta_table_name;
	}

	/**
	 * Returns the name of the primary key
	 *
	 * @return 	string	Name of primary key
	 *
	 * @since   1.0.0
	 */
	public static function getPrimaryKeyName() {
		return static::$primary_key;
	}

	/**
	 * Get Meta Table Primary Key name
	 * @return string
	 *
	 * @since   5.0.0
	 */
	public static function getMetaPrimaryKeyName() {
		return static::$meta_primary_key;
	}

	/**
	 * Returns table columns with primary key and created/updated columns
	 *
	 * @return array
	 *
	 * @since   5.0.0
	 */
	public static function getTableColumns() {
		$columns = static::$columns;

		if(!isset($columns[static::getPrimaryKeyName()])) {
			$primary_key = array(
				static::getPrimaryKeyName() => array(
					'type' => 'int',
					'unsigned' => true,
					'not_null' => true,
				)
			);

			$columns = array_merge($primary_key, $columns);
		}

		$columns = array_merge($columns, static::$columns_extra);

		return $columns;
	}

	/**
	 * Creates a filter name
	 *
	 * @param $function
	 * @param $action
	 * @return string
	 *
	 * @since 5.0.0
	 */
	public static function getFilterName($function, $action) {
		$slug = static::getSlug();

		$filter_name = "mtt-plugin/model/$slug/$function/$action";

		return $filter_name;
	}

	/**
	 * Get the primary key
	 *
	 * @return mixed
	 *
	 * @since    3.0.3
	 */
	public function getPrimaryKey() {
		return $this->{static::getPrimaryKeyName()};
	}

	// HELPERS
	/**
	 * Delete in database
	 *
	 * @return	bool	True if succesed, false if failed
	 *
	 * @since    1.0.0
	 */
	public function delete() {
		return $this->deleteByID($this->getPrimaryKey());
	}

	/**
	 * Save the current instance to the database
	 *
	 * @return bool
	 *
	 * @since    1.0.0
	 */
	public function save() {
		$columns = get_object_vars($this);

		$columns = \apply_filters($this->getFilterName('save', 'columns'), $columns, $this->{static::$primary_key});

		if(isset($this->{static::$primary_key}) && $this->{static::$primary_key} > 0) {
			$return = $this->update($this->{static::$primary_key}, $columns);
		} else {
			$return = $this->insert($columns);
			$this->{static::getPrimaryKeyName()} = $return->{static::getPrimaryKeyName()};
		}

		\do_action(static::$table_name . '_after_save', $columns, $this->{static::$primary_key});

		return $return;
	}

	/**
	 * Insert object in db
	 *
	 * @param array $columns
	 *
	 * @return bool|static
	 *
	 * @since    1.0.0
	 */
	public static function insert(array $columns = array()) {
		global $wpdb;

		$columns['created_date'] = \current_time('mysql');
		$columns['created_by'] = \get_current_user_id();

		foreach($columns as $column => $data) {
			if(!self::tableColumnExists($column)) {
				unset($columns[$column]);
			}
		}

		if(!$wpdb->insert(static::getTableName(), $columns)) {
			return false;
		};

		$columns[static::getPrimaryKeyName()] = $wpdb->insert_id;

		\do_action(static::$table_name . '_after_insert', $columns);

		return new static($columns);
	}

	/**
	 * Check if table column exists
	 *
	 * @param $column_name
	 *
	 * @return bool
	 */
	public static function tableColumnExists($column_name) {
		$columns = static::getTableColumns();

		return isset($columns[$column_name]);
	}

	/**
	 * Update object in database
	 *
	 * @param       $id
	 * @param array $columns
	 *
	 * @return bool
	 *
	 * @since    1.0.0
	 */
	public static function update($id, $columns = array()) {
		global $wpdb;

		$columns['updated_date'] = \current_time('mysql');
		$columns['updated_by'] = \get_current_user_id();

		$columns = \apply_filters(static::getFilterName('update', 'columns'), $columns, $id);

		foreach($columns as $column => $data) {
			if(!isset(static::$columns[$column])) {
				unset($columns[$column]);
			}
		}

		if(!$wpdb->update(static::getTableName(), $columns, array(static::$primary_key  => $id))) {
			return false;
		};

		\do_action(static::$table_name . '_after_update', $columns, $id);

		return true;
	}

	/**
	 * Check if table needs to be updated
	 *
	 * @since	1.0.0
	 */
	public static function updateTable() {
		static::createTable();
		static::createMetaTable();
	}

	/**
	 * Creates or update table
	 *
	 * @since    1.0.0
	 */
	public static function createTable() {
		static::createDBTable(static::getTableName(), static::getTableColumns(), static::getPrimaryKeyName(), static::$db_version);
	}

    /**
     * Creates Meta Table
     *
     * @since   5.0.0
     */
	public static function createMetaTable() {
		static::createDBTable(static::getMetaTableName(), static::$meta_columns, static::getMetaPrimaryKeyName(), static::$db_version);
	}

    /**
     * Creates Table - Rewritten and sepperated from createTable
     *
     * @param $table_name
     * @param $columns
     * @param $primary_key
     * @param $version
     *
     * @since   5.0.0
     */
	public static function createDBTable($table_name, $columns, $primary_key, $version) {
		if(empty($table_name) || $version == \get_option($table_name . '_db_version')) {
			return;
		}

		global $wpdb;

		require_once(ABSPATH . 'wp-admin/includes/upgrade.php');

		$charset = $wpdb->get_charset_collate();

		$sql_columns = "";

		foreach($columns as $name => $column) {
			$sql_columns .= $name . " " . $column['type'] . (isset($column['unsigned']) && $column['unsigned'] ? " unsigned" : "") . (isset($column['default']) ? " DEFAULT '" . $column['default'] . "'" : "") . (isset($column['not_null']) && $column['not_null'] ? " NOT NULL" : "") . ($name === static::$primary_key ? " AUTO_INCREMENT" : "") . ", 
			";
		}

		$sql = "CREATE TABLE $table_name (
			$sql_columns
			PRIMARY KEY  ($primary_key),
			UNIQUE KEY $primary_key ($primary_key)
		) $charset;";

		\dbDelta($sql);

		\update_option($table_name . '_db_version', $version);
	}

	/**
	 * Drops table
	 *
	 * @since    1.0.0
	 */
	public static function dropTable() {
		global $wpdb;

		$wpdb->query("DROP TABLE IF EXISTS " . static::getTableName());

		\delete_option(static::$table_name . '_db_version');
	}

	/**
	 * Helper function to perform SQL query
	 *
	 * @param        $sql
	 * @param bool   $single
	 * @param string $return
	 *
	 * @return static
	 *
	 * @since    1.0.0
	 */
	protected static function performSQL($sql, $single = false, $return = 'OBJECT') {
		global $wpdb;

		if($single) {
			$results = $wpdb->get_row($sql, 'ARRAY_A');

			if($results !== NULL) {
				if($return != 'ARRAY_A') {
					$results = new static($results);
				}
			}
		} else {
			$results = $wpdb->get_results($sql, 'ARRAY_A');

			if($return != 'ARRAY_A') {
				foreach($results as $key => $result) {
					$results[$key] = new static($result);
				}
			}
		}

		return $results;
	}

	/**
	 * Get obejcts from databse using raw sql statement
	 *
	 * @param	string 	$sql	SQL statement
	 * @param 	bool 	$cache	Should WordPress get the result from the cache if it exists
	 *
	 * @return 	array	Array of objects
	 *
	 * @since    1.0.0
	 */
	public static function getRaw($sql, $cache = true, $single = false) {
		if($single && !strpos($sql, 'LIMIT')) {
			$sql = trim($sql) . ' LIMIT 1';
		}

		$cache_key = md5(static::$table_name . $sql);

		$results = false;

		if($cache) {
			if(array_key_exists($cache_key, static::$cache)) {
				$results = static::$cache[$cache_key];
			} else {
				$results = \wp_cache_get($cache_key);
			}
		}

		if(!$results) {
			$results = static::performSQL($sql, $single);

			\wp_cache_set($cache_key, $results);
		}

		static::$cache[$cache_key] = $results;

		return $results;
	}

	/**
	 * Get single object from data base using raw sql
	 *
	 * @param      $sql
	 * @param bool $cache
	 *
	 * @return array
	 *
	 * @since    1.0.0
	 */
	public static function getFirstRaw($sql, $cache = true) {
		return static::getRaw($sql, $cache, true);
	}

	/**
	 * Get objects from database using arguments
	 *
	 * @param 	array 	$args	Arguments to use for the sql query
	 * @param 	bool  	$cache	Should WordPress get the result from the cache if it exists
	 *
	 * @return array
	 *
	 * @since    1.0.0
	 */
	public static function get(array $args = array(), $cache = true, $single = false) {
		$defaults = array(
			'posts_per_page' => $single ? 1 : \get_option('posts_per_page'),
			'offset' => 0,
			'page' => 1,
			'orderby' => static::$primary_key,
			'order' => 'ASC',
			'where' => array(),
			'columns' => '*',
			'with' => array(),
			'return' => 'OBJECT',
			'joins' => '',
			'as' => '',
		);

		$args = \wp_parse_args($args, $defaults);

		$cache_key = md5(static::$table_name . serialize($args));

		$results = false;

		if($cache) {
			if(array_key_exists($cache_key, static::$cache) && array_key_exists($cache_key . '_count_total', static::$cache)) {
				$results = static::$cache[$cache_key];
				static::$total_num_rows = static::$cache[$cache_key . '_count_total'];
			} else {
				$results = \wp_cache_get($cache_key);
				static::$total_num_rows = \wp_cache_get($cache_key . '_count_total');
			}
		}

		if(!$results) {
			global $wpdb;

			$select = is_array($args['columns']) ? \implode(', ', $args['columns']) : $args['columns'];

			$from = static::getTableName();

			$as = empty($args['as']) ? '' : 'AS ' . $args['as'];

			$where = static::createWhereSQLPart($args['where']);

			$offset = $args['offset'] + (($args['page'] * $args['posts_per_page']) - $args['posts_per_page']);

			$post_per_page = $args['posts_per_page'] > 0 ? $args['posts_per_page'] : PHP_INT_MAX;

			$sql = \implode(' ', array(
				'SELECT',
				'SQL_CALC_FOUND_ROWS',
				$select,
				'FROM',
				$from,
				$as,
				$args['joins'],
				$where,
				'ORDER BY',
				$args['orderby'],
				$args['order'],
				'LIMIT',
				$post_per_page,
				'OFFSET',
				$offset,
			));

			$results = static::performSQL($sql, $single, $args['return']);

			static::$total_num_rows = $wpdb->get_var('SELECT FOUND_ROWS()');

			\wp_cache_set($cache_key . '_count_total', static::$total_num_rows);

			\wp_cache_set($cache_key, $results);
		}

		static::$cache[$cache_key] = $results;
		static::$cache[$cache_key . '_count_total'] = static::$total_num_rows;

		return $results;
	}

	/**
	 * Get all
	 *
	 * @param array $args
	 * @param bool  $cache
	 *
	 * @return array
	 *
	 * @since    3.0.3
	 */
	public static function getAll(array $args = array(), $cache = true) {
		$defaults = array(
			'posts_per_page' => -1,
		);

		$args = \wp_parse_args($args, $defaults);

		return static::get($args);
	}

	/**
	 * Get by column
	 *
	 * @param       $column
	 * @param       $value
	 * @param array $args
	 * @param bool  $cache
	 * @param bool  $single
	 *
	 * @return array
	 *
	 * @since    3.0.3
	 */
	public static function getBy($column, $value, array $args = array(), $cache = true, $single = false) {
		$defaults = array(
			'where' => array(
				$column	=> $value,
			)
		);

		$args = \wp_parse_args($args, $defaults);

		return static::get($args, $cache, $single);
	}

	/**
	 * Get by primary key
	 *
	 * @param       $value
	 * @param array $args
	 * @param bool  $cache
	 *
	 * @return array
	 *
	 * @since    3.0.3
	 */
	public static function getByPrimaryKey($value, array $args = array(), $cache = true) {
		return static::getBy(static::getPrimaryKeyName(), $value, $args, $cache, true);
	}

	/**
	 * Get single object form data base
	 *
	 * @param 	array	$args
	 * @param	bool	$cache
	 *
	 * @return	array
	 *
	 * @since   1.0.0
	 */
	public static function getFirst(array $args = array(), $cache = true) {
		return static::get($args, $cache, true);
	}

	/**
	 * Returns the total number of objects from last query (without LIMIT)
	 *
	 * @return int
	 *
	 * @since    1.0.0
	 */
	public static function getTotalCount() {
		return static::$total_num_rows;
	}

	/**
	 * Creates the where part of the sql query
	 *
	 * @param 	array	$pieces		Pieces $column => $value, array($column, $value) or array($column, $operator, $value)
	 *
	 * @return 	string
	 *
	 * @since   1.0.0
	 */
	protected static function createWhereSQLPart($pieces = array(), $operator = 'AND') {
		if(!is_array($pieces)) {
			return $pieces;
		}

		if(empty($pieces)) {
			return '';
		}

		$str = 'WHERE';

		$str .= static::createWhereSQLPartPiece($pieces, $operator);

		return $str;
	}

	/**
	 * Creates every piece of the where sql part
	 * @param        $pieces
	 * @param string $operator
	 *
	 * @return string
	 *
	 * @since    3.0.3
	 */
	protected static function createWhereSQLPartPiece($pieces, $operator = 'AND') {
		if(!is_array($pieces)) {
			return $pieces;
		}
		if(empty($pieces)) {
			return '';
		}

		global $wpdb;

		if(isset($pieces['relation'])) {
			$operator = $pieces['relation'];
			unset($pieces['relation']);
		}

		$str = ' (';

		$comparisons = array();

		foreach($pieces as $key => $piece) {
			if(is_string($piece) && is_int($key)) {
				$comparisons[] = $piece;
			} elseif(is_array($piece) && isset($piece['relation'])) {
				$comparisons[] = static::createWhereSQLPartPiece($piece, $piece['relation']);
			} else {
				$compare_operator = "=";

				if(is_array($piece)) {
					if(count($piece) === 2) {
						$column = $piece[0];
						$value = $piece[1];
					} else if(count($piece) === 3) {
						$column = $piece[0];
						$compare_operator = $piece[1];
						$value = $piece[2];
					}
				} else {
					$column = $key;
					$value = $piece;
				}

				if($compare_operator == 'IN' || $compare_operator == 'NOT IN' && is_array($value)) {
					$value = '(' . implode(',', $value) . ')';
					$comparisons[] = "$column $compare_operator " . $value;
				} else {
					if(is_int($value)) {
						$value_type = "%d";
					} else if(is_float($value)) {
						$value_type = "%s";
					} else {
						$value_type = "%s";
					}
					$sql = "$column $compare_operator $value_type";

					$comparisons[] = $wpdb->prepare($sql, $value);
				}
			}
		}

		$str .= \implode($comparisons, ' ' . $operator . ' ');

		$str .= ')';

		return $str;
	}

	/**
	 * Delete row by id
	 *
	 * @param 	$id
	 *
	 * @return	bool	ID of row to delete
	 *
	 * @since   1.0.0
	 */
	public static function deleteByID($id) {
		global $wpdb;

		if(!$wpdb->delete(static::getTableName(), array(static::$primary_key => $id))) {
			return false;
		}

		\do_action(static::$table_name . '_after_delete', $id);

		return true;
	}

	/**
	 * Delete rows where coloumns = value
	 *
	 * @param 	array 	$columns	Column => value to match
	 *
	 * @return	mixed
	 *
	 * @since   1.0.0
	 */
	public static function deleteWhere(array $columns = array()) {
		global $wpdb;

		return $wpdb->delete(static::getTableName(), $columns);
	}

	/**
	 * Check if object exists
	 *
	 * @param array $columns
	 *
	 * @return array|int
	 *
	 * @since    1.0.0
	 */
	public static function exists(array $columns = array(), $except = array()) {
		global $wpdb;

		if(empty($columns)) {
			return 0;
		}

		$sql = "SELECT " . static::$primary_key . " FROM " . static::getTableName() . " WHERE ";

		foreach($columns as $key => $value) {
			$sql .= $wpdb->prepare("$key = %s && ", $value);
		}

		foreach($except as $key => $value) {
			$sql .= $wpdb->prepare("$key != %s && ", $value);
		}

		$sql = trim($sql, ' \t\n\r\0\x0B&');

		return static::getRaw($sql);
	}
}