<?php namespace MTTPlugin\Lib\Abstracts;

use MTTPlugin\Lib\Tools\Loader;

abstract class Plugin {

	use \MTTPlugin\Lib\Traits\Translate;
	use \MTTPlugin\Lib\Traits\Templating;

	/**
	 * URL to update file
	 */
	var $update_url = 'https://app.re-zip.dk/plugin/woocommerce.json?hash=63q6cxwZKd6Qramj';

	/**
	 * Plugin info - Should be declared in child class
	 */
	protected static $plugin_info = array();

	/**
	 * Plugin constructor
	 *
	 * @param $plugin_file      Path to plugin file
	 * @param $plugin_namespace Namespace of plugin
	 */
	public function __construct( $plugin_file, $plugin_namespace ) {

        if ( isset( $_SERVER['HTTP_HOST'] ) && $_SERVER['HTTP_HOST'] == 'rezip.1902dev1.com' ) {
            $this->update_url = 'https://rezip.1902dev1.com/plugin/woocommerce.json?hash=63q6cxwZKd6Qramj';
        }

		$this->setPluginInfo('plugin', $this);
		$this->setPluginInfo('file', $plugin_file);
		$this->setPluginInfo('root', \plugin_dir_path($plugin_file));
		$this->setPluginInfo('namespace', $plugin_namespace);
		$this->setPluginInfo('data', static::readPluginData($plugin_file));

        $this->loadPluginUpdater();

		if(is_admin()) {
			$admin_class = $plugin_namespace . '\Plugin\PluginAdmin';
			$this->setPluginInfo('admin', new $admin_class($this));
		}

		$this->registerFrameworkActions();
		$this->registerFrameworkFilters();

		$this->registerActions();
		$this->registerFilters();

		$this->registerControllers();
		$this->registerModels();
		$this->registerCustomPostTypes();

		\do_action('after_plugin_loaded_' . static::getTextDomain(), $this);
	}

    // SETTERS

	/**
	 * Set plugin info
	 *
	 * @param $name
	 * @param $info
	 */
    protected function setPluginInfo($name, $info) {
        static::$plugin_info[$name] = $info;
    }


    // GETTERS

	/**
	 * Get plugin root path
	 *
	 * @param string $path Path relative to root
	 *
	 * @return string
	 */
    public static function getRoot($path = '') {
        return static::getPluginInfo('root') . static::decodeFolderStructure($path);
    }

	/**
	 * Alias for getRoot
	 *
	 * @param string $path Path relative to root
	 *
	 * @return string
	 */
    public static function getPluginRoot($path = '') {
        return self::getRoot($path);
    }

	/**
	 * Get plugin version
	 *
	 * @return mixed|string
	 */
    public static function getVersion() {
        return static::getPluginData('Version');
    }

	/**
	 * Get plugin text domain
	 *
	 * @return mixed|string
	 */
    public static function getTextDomain() {
        return static::getPluginData('TextDomain');
    }

	/**
	 * Get plugin info
	 *
	 * @param $name
	 *
	 * @return mixed|string
	 */
	public static function getPluginInfo($name) {
		if(isset(static::$plugin_info[$name])) {
			return static::$plugin_info[$name];
		}

		return '';
	}

	/**
	 * Get path to plugin file
	 *
	 * @return mixed|string
	 */
    public static function getPluginFilePath() {
        return static::getPluginInfo('file');
    }

	/**
	 * Get URL to plugin root
	 *
	 * @param string $url
	 *
	 * @return string
	 */
    public static function getPluginRootURL($url = '') {
        return \plugin_dir_url(static::getPluginFilePath()) . $url;
    }

	/**
	 * Get plugin data
	 *
	 * @param $name
	 *
	 * @return mixed|string
	 */
    public static function getPluginData($name) {
        $data = static::getPluginInfo('data');

        if(is_array($data) && isset($data[$name])) {
            return $data[$name];
        }

        return '';
    }


    // HELPERS

	/**
	 * Read Plugin Data from Plugin file
	 *
	 * @param $plugin_file
	 *
	 * @return mixed
	 */
	protected static function readPluginData($plugin_file) {
		require_once(ABSPATH . 'wp-admin/includes/plugin.php');

		return \get_plugin_data($plugin_file, false, false);
	}

	/**
	 * Load plugin updater if exists
	 */
	public function loadPluginUpdater() {
		if(!empty($this->update_url)) {
			if(!class_exists('PucFactory')) {
				require_once($this->getRoot('lib/vendor/plugin-update-checker/plugin-update-checker.php'));
			}

			\PucFactory::buildUpdateChecker(
				$this->update_url,
				$this->getPluginFilePath(),
				$this->getTextDomain()
			);
		}
	}

	/**
	 * Register framework actions
	 */
	public function registerFrameWorkActions() {
		\register_activation_hook($this->getPluginFilePath(), array($this, 'doActivationHook'));
		\register_deactivation_hook($this->getPluginFilePath(), array($this, 'doDeactivationHook'));
		\register_uninstall_hook($this->getPluginFilePath(), array(__CLASS__, 'doUninstallHook'));
	}

	/**
	 * Register framework filters
	 */
	public function registerFrameworkFilters() {
	    Loader::addFilter(static::getFilterName('args', 'templating'), $this, 'addArgsToTemplates', 1, 3);
	}

	/**
	 * Register Actions
	 */
	public function registerActions() {

	}

	/**
	 * Register Filters
	 */
	public function registerFilters() {

	}

	/**
	 * Activation hook
	 */
	public function doActivationHook() {
		update_option( 'rezip-type', 'fee' );
	}

	/**
	 * Deactivation hook
	 */
	public function doDeactivationHook() {

	}

	/**
	 * Uninstall hook
	 */
	public static function doUninstallHook() {

	}

	/**
	 * Register Controllers
	 */
	public function registerControllers() {
		$this->registerClasses('plugin.controllers', 'controller', 'Controller');
	}

	/**
	 * Register Models
	 */
	public function registerModels() {
		$this->registerClasses('plugin.models', 'class', '', 'register');
	}

	/**
	 * Register Custom Post Types
	 */
	public function registerCustomPostTypes() {
		$this->registerClasses('plugin.post-types', 'class');
	}

	/**
	 * Helper function to register classes in folder
	 *
	 *
	 */
	private function registerClasses($dir, $type = '', $append = '', $call = '__construct') {

		$_dir = static::getRoot($dir);

		if(!is_dir($_dir)) {
			return;
		}

		$dir_iterator = new \RecursiveDirectoryIterator($_dir);

		foreach($dir_iterator as $file) {
			$name = $file->getFileName();

			if(!empty($type) && substr($name, 0, strlen($type) + 1) != $type . '.') {
				continue;
			}
			$class_name_parts = explode('-', substr(substr($name,strlen($type) + 1),0,-4));

			$class_name = '';
			foreach($class_name_parts as $part) {
				$class_name .= ucfirst($part);
			}

			$folder_namespace = $this->getPluginInfo('namespace');

			$dir = static::decodeFolderStructure($dir);
			$dir_parts = explode('/', $dir);

			foreach($dir_parts as $part) {
				$name_parts = explode('-', $part);
				$folder_namespace .= '\\';
				foreach($name_parts as $name_part) {
					$folder_namespace .= ucfirst($name_part);
				}
			}

			$class_name = $folder_namespace . '\\' . $class_name . $append;

			if(class_exists($class_name)) {
				if($call == '__construct') {
					new $class_name();
				} else {
					$class_name::{$call}();
				}
			}
		}
	}


	/**
	 * Includes standard args in templates
	 *
	 * @param $args
	 * @param $template_name
	 * @param $file
	 *
	 * @return mixed
	 */
	public function addArgsToTemplates($args, $template_name, $file) {
		$args['plugin'] = static::getPluginInfo('plugin');

		if(\is_admin()) {
			$args['plugin_admin'] = $args['plugin']->getPluginInfo('admin');
		}

		return $args;
	}
}