<?php namespace MTTPlugin\Lib\Abstracts;

use MTTPlugin\Lib\Tools\Loader;

abstract class PostType {

	protected static $post_type_info = array();

	public function __construct() {
		$this->initArgs();
		$this->registerHooks();

		return $this;
	}

	/**
	 * Overwriting purpose - Init Post Type Arguments
	 * @return mixed
	 */
	abstract protected function initArgs();

	/**
	 * Set post type information (use key/value array or set single with key and value)
	 * @param $key_or_array (key or key/value array)
	 * @param null $value value
	 */
	protected static function setPostTypeInfo($key_or_array, $value = null) {
		if(!is_array($key_or_array)) {
			$key_or_array = array($key_or_array => $value);
		}

		static::$post_type_info = array_merge((array) static::$post_type_info, $key_or_array);
	}

	/**
	 * Get post type information - Set key to null returns all post type informations as an array
	 * @param null $key Key of information og null for all information
	 * @param null $default Default value to return if no information found
	 * @return array|mixed|null
	 */
	protected static function getPostTypeInfo($key = null, $default = null) {
		if(is_null($key)) {
			return static::$post_type_info;
		}

		if(isset(static::$post_type_info[$key])) {
			return static::$post_type_info[$key];
		}

		return $default;
	}

	/**
	 * Get post type slug
	 * @return array|mixed|null
	 */
	protected static function getSlug() {
		return static::getPostTypeInfo('slug');
	}

	/**
	 * Get args for post type registration
	 * @return array|mixed|null
	 */
	protected static function getArgs() {
		return static::getPostTypeInfo('args');
	}

	/**
	 * Set args for post type registration
	 * @param $args
	 */
	protected static function setArgs($args) {
		$args = wp_parse_args($args, static::getArgs());

		static::setPostTypeInfo('args', $args);
	}

	/**
	 * Set arg for post type registration
	 * @param $key
	 * @param $arg
	 */
	protected static function setArg($key, $arg) {
		static::setArgs(array(
			$key => $arg
		));
	}

	/**
	 * Get arg for post type registration
	 * @param $key
	 * @param bool $default
	 * @return bool
	 */
	protected function getArg($key, $default = false) {
		$args = self::getArgs();

		if(isset($args[$key])) {
			return $args[$key];
		}

		return $default;
	}

	/**
	 * Set support args (tile, editor, author, thumbnail, excerpt, trackbacks, custom-fields, comments, revisions, page-attributes, post-formats)
	 * @param string|array $supports
	 */
	protected function setSupports($supports = array('title', 'editor', 'thumbnail')) {
		static::setArg('supports', (array) $supports);
	}

	/**
	 * Set visibility of post type
	 * @param string $visibility
	 */
	protected function setVisibility($visibility = 'puplic') {
		$args = array();

		switch($visibility) {
			case 'public':
				$args['public'] = true;
				$args['has_archive'] = true;
				break;
			case 'admin':
				$args['public'] = false;
				$args['show_ui'] = true;
				break;
			default:
				$args['public'] = false;
		}

		static::setArgs($args);
	}

	/**
	 * Set slug and name of post type
	 * @param $slug Slug
	 * @param $plural_name Plural Name
	 * @param $singular_name Singular Name
	 */
	protected function setName($slug, $plural_name, $singular_name) {
		static::setPostTypeInfo('slug', $slug);

		$args = array(
			'label' => $plural_name,
			'labels' => array(
				'name' => $plural_name,
				'singular_name' => $singular_name,
			),
		);

		static::setArgs($args);
	}

	/**
	 * Set Icon for admin menu (dashicon or SVG)
	 * @param $icon
	 */
	protected function setIcon($icon) {
		static::setArg('menu_icon', $icon);
	}

	/**
	 * Set menu position
	 * @param $position
	 * @param bool $menu_separator
	 */
	protected function setMenuPosition($position, $menu_separator = false) {
		static::setArg('menu_position', $position);
		static::setPostTypeInfo('menu_separator', $menu_separator);
	}

	/**
	 * Registers a custom admin column
	 * @param $slug
	 * @param $callback
	 */
	public function addCustomAdminColumn($slug, $title, $component, $callback, $sortable_component = null, $sortable_callback = null) {
		$custom_admin_columns = array_merge(static::getPostTypeInfo('custom_admin_columns', array()), array($slug => array('slug' => $slug, 'title' => $title, 'component' => $component, 'callback' => $callback, 'sortable_component' => $sortable_component, 'sortable_callback' => $sortable_callback)));

		static::setPostTypeInfo('custom_admin_columns', $custom_admin_columns);
	}

	/**
	 * Add a taxonomy associated with the post type
	 * @param $slug
	 * @param $name
	 * @param array $args
	 */
	protected function addTaxonomy($slug, $name, $args = array()) {
		$taxonomies = static::getPostTypeInfo('taxonomies');
		if(!isset($taxonomies[$slug])) {
			$defauls = array(
				'label' => $name,
				'labels' => array(
					'name' => $name
				),
				'hierarchical' => true,
			);

			$defauls = wp_parse_args($defauls, static::getArgs());

			static::setPostTypeInfo('taxonomies', array_merge((array) $taxonomies, array($slug => wp_parse_args($args, $defauls))));
		}
	}

	/**
	 * Register Hooks
	 */
	protected function registerHooks() {
		Loader::addAction('init', $this, 'registerPostType', 10);
		Loader::addAction('init', $this, 'registerTaxonomies', 20);
		Loader::addAction('admin_menu', $this, 'unsetAddNew');
		Loader::addAction('admin_menu', $this, 'setMenuSeperator');
		Loader::addFilter('manage_' . static::getSlug() . '_posts_columns', $this, 'manageCustomAdminColumns');
		Loader::addAction('manage_' . static::getSlug() . '_posts_custom_column', $this, 'manageCustomAdminColumnsContent', 10, 2);
		Loader::addFilter('manage_edit-' . static::getSlug() . '_sortable_columns', $this, 'manageCustomAdminSortableColumns');
		Loader::addAction('pre_get_posts', $this, 'manageCustomAdminSortableColumnsCallback');
	}

	public function manageCustomAdminSortableColumnsCallback($query) {
		if(!is_admin()) {
			return;
		}

		$orderby = $query->get('orderby');

		$custom_admin_columns = static::getPostTypeInfo('custom_admin_columns', array());

		if(isset($custom_admin_columns[$orderby])) {
			$column = $custom_admin_columns[$orderby];
			$callback = $column['sortable_component'] !== null ? array($column['sortable_component'], $column['sortable_callback']) : $column['sortable_callback'];
			if(!is_null($callback)) {
				call_user_func_array($callback, array($query));
			}
		}
	}

	/**
	 * manage sortable columns
	 *
	 * @param $columns
	 *
	 * @return mixed
	 */
	public function manageCustomAdminSortableColumns($columns) {
		$custom_admin_columns = static::getPostTypeInfo('custom_admin_columns', array());
		foreach($custom_admin_columns as $column) {
			$callback = $column['sortable_component'] !== null ? array($column['sortable_component'], $column['sortable_callback']) : $column['sortable_callback'];
			if(!is_null($callback)) {
				$columns[$column['slug']] = $column['slug'];
			}
		}
		return  $columns;
	}

	/**
	 * Manages content callbacks for custom admin columns
	 * @param $column_name
	 * @param $post_id
	 */
	public function manageCustomAdminColumnsContent($column_name, $post_id) {
		$custom_admin_columns = static::getPostTypeInfo('custom_admin_columns', array());
		if(isset($custom_admin_columns[$column_name])) {
			$column = $custom_admin_columns[$column_name];
			$callback = $column['component'] !== null ? array($column['component'], $column['callback']) : $column['callback'];
			call_user_func_array($callback, array($column_name, $post_id));
		}
	}

	/**
	 * Registers custom columns in admin table for the post type
	 * @param $columns
	 * @return array
	 */
	public function manageCustomAdminColumns($columns) {
		$custom_admin_columns = static::getPostTypeInfo('custom_admin_columns', array());
		foreach($custom_admin_columns as $column) {
			$columns[$column['slug']] = $column['title'];
		}
		return  $columns;
	}

	/**
	 * Unset "Add New" menu item for post type
	 */
	public function unsetAddNew() {
		if(static::getPostTypeInfo('remove_add_new_menu_item', false)) {
			global $submenu;

			if(isset($submenu['edit.php?post_type=' . static::getSlug()]) && isset($submenu['edit.php?post_type=' . static::getSlug()][10])) {
				unset($submenu['edit.php?post_type=' . static::getSlug()][10]);
			}
		}
	}

	/**
	 * Sets menu separator before menu position
	 */
	public function setMenuSeperator() {
		if(static::getPostTypeInfo('menu_separator', false)) {
			global $menu;

			$menu[(int) static::getArg('menu_position') - 1] = array(
				'',
				'read',
				'separator29',
				'',
				'wp-menu-separator',
			);
		}
	}

	/**
	 * Registers post type
	 */
	public function registerPostType() {
		register_post_type(
			static::getSlug(),
			static::getArgs()
		);
	}

	/**
	 * Registers taxonomies
	 */
	public function registerTaxonomies() {
		foreach(static::getPostTypeInfo('taxonomies', array()) as $key => $args) {
			register_taxonomy($key, static::getSlug(), $args);
		}
	}

	/**
	 * Get all posts
	 * @param array $args
	 * @return mixed
	 */
	public static function getAll($args = array()) {
		return static::get(-1, $args);
	}

	/**
	 * Get Posts
	 * @param null $number_of_posts Null = default, -1 = all
	 * @param array $args
	 * @return mixed
	 */
	public static function get($number_of_posts = null, $args = array()) {
		$defaults = array(
			'post_type' => static::getSlug(),
		);

		if(!is_null($number_of_posts)) {
			$defaults['posts_per_page'] = $number_of_posts;
		}

		$args = wp_parse_args($args, $defaults);

		return get_posts($args);
	}

	/**
	 * Get post by ID(s)
	 * @param $ids (id or array of ids)
	 * @param array $args
	 * @return mixed
	 */
	public static function getByID($ids, $args = array()) {
		$defaults = array(
			'post__in' => (array) $ids
		);

		$args = wp_parse_args($args, $defaults);

		return static::get(count($args['post__in']), $args);
	}
}
