<?php namespace MTTPlugin\Lib\Classes;

use MTTPlugin\Lib\Tools\Loader;

/**
 * This class helps with URL rewriting (routes)
 *
 * @since 3.0.5
 */
class Route {

    protected static $routes = array();
    protected static $active_route = null;

    protected $context;
    protected $page_title = '';
    protected $template = null;
    protected $slug;
    protected $type;
    protected $route;
    protected $route_regex;
    protected $url;
    protected $priority;
    protected $callback;
    protected $vars = array();
    protected $query_vars = array();
    protected $parsed_vars = array();

    public function __construct($slug, $type, $route, $callback, $priority = 'bottom') {
        $this->slug = $slug;
        $this->type = $type;
        $this->route = $route;
        $this->callback = $callback;
        $this->priority = $priority;

        static::$routes[$slug] = $this;

        return $this;
    }

    public function initRoute() {
        $this->route_regex = \preg_replace_callback('/\/{([a-z0-9_]+)(\?)?}/', array($this, 'setVar'), $this->getRoute());

        Loader::addFilter('query_vars', $this, 'registerVars');
    }

    public function registerVars($vars) {
        foreach($this->getQueryVars() as $name =>$var) {
            $vars[] = $name;
        }

        return $vars;
    }

    private function setVar($matches) {
        $var_name = $matches[1];
        $optional = isset($matches[2]) && $matches[2] == '?' ? true : false;

        if(isset($this->vars[$var_name]) && isset($this->vars[$var_name]['pattern'])) {
            $regex = $this->vars[$var_name]['pattern'];
        } else {
            $regex = '[^/]+';
        }

        $this->vars[$matches[1]] = array(
            'name' => $matches[1],
            'optional' => $optional,
            'pattern' => $regex,
        );

        $return = "/($regex)";

        if($optional) {
            $return = "(?:/?$return)?";
        }

        return $return;
    }

    public function getSlug() {
        return $this->slug;
    }

    public function getVars() {
        if(empty($this->vars)) {
            $this->initRoute();
        }
        return $this->vars;
    }

    private function getQueryVars() {
        if(empty($this->query_vars)) {
            $i = 1;
            foreach($this->getVars() as $var) {
                $this->query_vars[$var['name']] = '$matches[' . $i . ']';
                $i++;
            }
            $this->query_vars['route'] = $this->getSlug();
        }

        return $this->query_vars;
    }

    public function getParsedVars($name = null, $default = '') {
        if(empty($this->parsed_vars)) {
            foreach($this->getVars() as $var) {
                $query_var = \get_query_var($var['name'], null);
                if(!empty($query_var)) {
                    $this->parsed_vars[$var['name']] = $query_var;
                }
            }
        }
        if(!is_null($name)) {
            if(isset($this->parsed_vars[$name])) {
                return $this->parsed_vars[$name];
            }
            return $default;
        }
        return $this->parsed_vars;
    }

    public function where($var_name, $regex) {
        $this->vars[$var_name]['pattern'] = $regex;
        return $this;
    }

    public function context($context) {
        $this->context = $context;
        return $this;
    }

    public function setPageTitle($title) {
        $this->page_title = $title;
        return $this;
    }

    public function setTemplate($template) {
    	$this->template = $template;
    	return $this;
    }

    public function registerRoute() {
        $this->initRoute();

        \add_rewrite_rule($this->getRouteRegex(), $this->getQueryVars(), $this->getPriority());
    }

    public function getRoute() {
        return $this->route;
    }

    protected function getRouteRegex() {
        if(!isset($this->route_regex)) {
            $this->initRoute();
        }
        return $this->route_regex;
    }

    protected function getPriority() {
        return $this->priority;
    }

    public function resolveCallback() {
        if(is_string($this->callback)) {
            $frags = explode('@', $this->callback);
            if(count($frags) === 2) {
                $className = $this->context->controllerNamespace . $frags[0];
                $method = $frags[1];
                if(class_exists($className)) {
                    $this->callback = array(new $className($this->context), $method);
                }
            }
        }

        if(is_callable($this->callback)){
            return $this->callback;
        }
    }

    public function execute() {
        ob_start();
        \call_user_func_array($this->resolveCallback(), $this->getParsedVars());
        return ob_get_clean();
    }

    public function getURL($args = array()) {
        $url_ending = $this->route;

        foreach($args as $key => $arg) {
            $url_ending = \preg_replace('/{' . $key . '(\?)?}/', $arg, $url_ending);
        }

        $url = get_home_url() . '/' . $url_ending;

        return $url;
    }

    public function getPageTitle() {
        if(is_callable($this->page_title)) {
            return call_user_func($this->page_title, $this);
        }
        return $this->page_title;
    }

    // STATIC FUNCTIONS

    public static function get($slug, $route, $callback, $priority = 'bottom') {
        return new static($slug, 'GET', $route, $callback, $priority);
    }

    public static function post($slug, $route, $callback, $priority = 'bottom') {
        return new static($slug, 'POST', $route, $callback, $priority);
    }

    public static function registerRoutes() {
        foreach(static::$routes as $route) {
            $route->registerRoute();
        }
    }

    public static function checkRoutes($preempt) {
        if(!is_null(static::getActiveRoute())) {
            $preempt = true;
        }
        return $preempt;
    }

    public static function init() {
        static::registerHooks();
        static::registerFilters();
    }

    public static function registerHooks() {
        Loader::addAction('init', __CLASS__, 'registerRoutes');
    }

    public static function registerFilters() {
        Loader::addFilter('pre_handle_404', __CLASS__, 'checkRoutes');
        Loader::addFilter('the_posts', __CLASS__, 'setupPostGhost', 10, 2);
        Loader::addFilter('the_content', __CLASS__, 'theContentFilter', 1);
	    Loader::addFilter('template_include', __CLASS__, 'changeTemplate');
    }

    public static function changeTemplate($template) {
	    global $wp_query;
    	$active_route = static::getActiveRoute();
		if(!is_null($active_route) && isset($wp_query->query['route']) && $active_route->getSlug() == $wp_query->query['route'] && !is_null($active_route->template)) {
			$template = $active_route->template;
		}

		return $template;
    }

    public static function setupPostGhost($posts, $wp_query) {
        $active_route = static::getActiveRoute();
        if(!is_null($active_route) && isset($wp_query->query['route']) && $active_route->getSlug() == $wp_query->query['route']) {
            $posts = array(
                new \WP_Post((object) array(
                    'ID' => -1,
                    'post_content' => $active_route->execute(),
                    'post_title' => $active_route->getPageTitle(),
                    'post_name' => $active_route->getSlug(),
                    'post_status' => 'static',
                    'comment_status' => 'closed',
                    'ping_status' => 'closed',
                    'comment_count' => 0,
                    'post_author' => 1,
                    'guid' => \get_home_url(null, $active_route->getSlug()),
                    'post_date' => \current_time('mysql'),
                    'post_date_gmt' => \current_time('mysql', 1),
                    'filter' => 'raw',
                ))
            );

            $wp_query->is_page = true;
            $wp_query->is_singular = true;
        }
        return $posts;
    }

    public static function getRouteURL($slug, $args = array()) {
        $route = static::getRouteDef($slug);

        return $route->getURL($args);
    }

    private static function getRouteDef($slug = '') {
        if(!empty($slug) && isset(static::$routes[$slug])) {
            return static::$routes[$slug];
        }
        return null;
    }

    public static function getActiveRoute() {
        if(is_null(static::$active_route)) {
            static::$active_route = static::getRouteDef(\get_query_var('route', null));
        }
        return static::$active_route;
    }

    public static function theContentFilter($content) {
        if(isset($GLOBALS['post']) && $GLOBALS['post']->ID === -1) {
            remove_filter('the_content', 'wpautop');
        }
        return $content;
    }
}